
/* memo.cc */

#include "memo.h"
#include "overview.h"

#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QHeaderView>
#include <QMenu>

#include <KLocalizedString>
#include <KIconLoader>

#include <ktexteditor/document.h>
#include <ktexteditor/view.h>

#include <interfaces/icore.h>
#include <interfaces/idocumentcontroller.h>

#include <language/duchain/declaration.h>
#include <language/duchain/duchainutils.h>
#include <language/duchain/duchainlock.h>
#include <language/duchain/duchain.h>

#include "trace.h"

// using namespace KDevelop;

/* ----------------------------------------------------------------------- */

namespace OverviewModule
{

/* ----------------------------------------------------------------------- */

MemoWidget::MemoWidget (QWidget* parent, OverviewPlugin* plugin) :
   QWidget (parent),
   m_plugin (plugin),
   m_tree (new QTreeWidget (this)),
   m_jumpToDeclaration (NULL),
   m_jumpToDefinition (NULL),
   m_jumpToInterface (NULL),
   m_jumpToImplementation (NULL)
{
   setObjectName ("MemoView");
   setWindowTitle (i18n ("Memo"));
   setWindowIcon (SmallIcon ("code-class"));

   // m_tree->header()->setResizeMode (0, QHeaderView::ResizeToContents);
   // m_tree->header()->setStretchLastSection (false);
   m_tree->header()->hide();
   m_tree->setIndentation(10);

   QVBoxLayout* vbox = new QVBoxLayout (this);
   vbox->setMargin (0);
   vbox->addWidget (m_tree);
   setLayout (vbox);

   setWhatsThis (i18n ("Memo"));

   connect (m_tree,
            SIGNAL (itemDoubleClicked (QTreeWidgetItem*, int)),
            this,
            SLOT (treeDoubleClick (QTreeWidgetItem*, int)));

   m_jumpToDeclaration = new KAction (this);
   m_jumpToDeclaration->setText (i18n ("Jump to Declaration"));
   m_jumpToDeclaration->setIcon (KIcon ("go-jump-declaration"));
   connect (m_jumpToDeclaration, SIGNAL (triggered (bool)), this, SLOT (jumpToDeclaration()));

   m_jumpToDefinition = new KAction (this);
   m_jumpToDefinition->setText (i18n ("Jump to Definition"));
   m_jumpToDefinition->setIcon (KIcon ("go-jump-definition"));
   connect (m_jumpToDefinition, SIGNAL (triggered (bool)), this, SLOT (jumpToDefinition()));

   m_jumpToInterface = new KAction (this);
   m_jumpToInterface->setText (i18n ("Jump to Interface"));
   connect (m_jumpToInterface, SIGNAL (triggered (bool)), this, SLOT (jumpToInterface()));

   m_jumpToImplementation = new KAction (this);
   m_jumpToImplementation->setText (i18n ("Jump to Implementation"));
   connect (m_jumpToImplementation, SIGNAL (triggered (bool)), this, SLOT (jumpToImplementation()));
}

MemoWidget::~MemoWidget()
{
}

/* ----------------------------------------------------------------------- */

void MemoWidget::contextMenuEvent (QContextMenuEvent* e)
{
   QMenu *menu = new QMenu (this);

   // Context* c;
   // QList<ContextMenuExtension> extensions = ICore::self()->pluginController()->queryPluginsForContextMenuExtensions( c );
   // ContextMenuExtension::populateMenu(menu, extensions);

   menu->addAction (m_jumpToDeclaration);
   menu->addAction (m_jumpToDefinition);
   menu->addAction (m_jumpToInterface);
   menu->addAction (m_jumpToImplementation);

   if (!menu->actions().isEmpty())
      menu->exec (QCursor::pos());
}

void MemoWidget::treeDoubleClick (QTreeWidgetItem* item, int column)
{
   // std::cout << std::endl;
   // std::cout << "DOUBLE CLICK " << item->text (0).data () << std::endl;
   // std::cout << std::endl;

   trace ("memo - tree double click");
   jumpToDeclaration ();
}

void MemoWidget::jumpToDeclaration ()
{
   trace ("memo - jump to declaration");
   QTreeWidgetItem * item = m_tree->currentItem ();
   if (item != NULL)
      trace ("memo - tree item: " + item->text (0));
}

void MemoWidget::jumpToDefinition ()
{
}

void MemoWidget::jumpToInterface ()
{
}

void MemoWidget::jumpToImplementation ()
{
}

/* ----------------------------------------------------------------------- */

static KDevelop::Declaration* cursorDeclaration()
{
   KDevelop::IDocument* doc = KDevelop::ICore::self()->documentController()->activeDocument();
   if (doc == NULL)
      return NULL;

   KTextEditor::Document* textDoc = doc->textDocument();
   if (textDoc == NULL)
      return NULL;

   KTextEditor::View* view = textDoc->activeView();
   if (view == NULL)
      return NULL;

   KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());

   return KDevelop::DUChainUtils::declarationForDefinition
             (KDevelop::DUChainUtils::itemUnderCursor (doc->url(), KDevelop::SimpleCursor (view->cursorPosition())));
}

/* ----------------------------------------------------------------------- */

void MemoWidget::display ()
{
   KDevelop::ICore::self()->uiController()->findToolView
      (i18n ("Memo"), m_plugin->m_memoFactory, KDevelop::IUiController::CreateAndRaise);

   KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
   KDevelop::Declaration * decl = cursorDeclaration();

   if (decl != NULL)
   {
      QTreeWidgetItem * item = new QTreeWidgetItem ();
      item->setText (0, decl->toString ());
      item->setIcon (0, KDevelop::DUChainUtils::iconForDeclaration (decl));
      m_tree->addTopLevelItem (item);

      KDevelop::DUContext * context = decl->internalContext ();
      if (context != NULL)
      {
         QVector <KDevelop::Declaration*> list = context->localDeclarations ();
         foreach (const KDevelop::Declaration * sub, list)
         {
            QTreeWidgetItem * item2 = new QTreeWidgetItem (item);
            item2->setText (0, sub->toString ());
            item2->setIcon (0, KDevelop::DUChainUtils::iconForDeclaration (sub));
            item2->setToolTip (0, sub->comment ());
         }
      }
   }
   else
   {
      QTreeWidgetItem * item = new QTreeWidgetItem ();
      item->setText (0, "none");
      m_tree->addTopLevelItem (item);
   }

   lock.unlock ();
}

/* ----------------------------------------------------------------------- */

} // end of namespace


#include "memo.moc"
