
/* navigator.cc */

#include "navigator.h"
#include "overview.h"

#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QHeaderView>
#include <QMenu>

#include <KLocalizedString>
#include <KIconLoader>

#include <ktexteditor/document.h>

#include <interfaces/icore.h>
#include <interfaces/idocumentcontroller.h>

#include <language/duchain/duchainutils.h>
#include <language/duchain/duchainlock.h>
#include <language/duchain/duchain.h>
#include <language/duchain/declaration.h>
#include <language/duchain/functiondefinition.h>

#include <iostream>
#include <typeinfo>
#include "trace.h"

// using namespace KDevelop;

/* ----------------------------------------------------------------------- */

namespace OverviewModule
{

/* ----------------------------------------------------------------------- */

NavigatorWidget::NavigatorWidget (QWidget* parent, OverviewPlugin* plugin) :
      QWidget (parent),
      m_plugin (plugin),
      m_tree (new QTreeWidget (this)),
      m_jumpToSource (NULL),
      m_jumpToDeclaration (NULL),
      m_jumpToDefinition (NULL),
      m_jumpToInterface (NULL),
      m_jumpToImplementation (NULL)
{
   setObjectName ("NavigatorView");
   setWindowTitle (i18n ("Navigator"));
   setWindowIcon (SmallIcon ("code-class"));

   m_tree->header()->setResizeMode (0, QHeaderView::ResizeToContents);
   // m_tree->header()->setStretchLastSection (false);
   m_tree->header()->hide();
   m_tree->setColumnCount (1);

   QVBoxLayout* vbox = new QVBoxLayout (this);
   vbox->setMargin (0);
   vbox->addWidget (m_tree);
   setLayout (vbox);

   setWhatsThis (i18n ("Navigator"));

   connect (KDevelop::ICore::self()->documentController(),
            SIGNAL (documentClosed (KDevelop::IDocument*)),
            this,
            SLOT (documentClosed (KDevelop::IDocument*)));

   connect (KDevelop::ICore::self()->documentController(),
            SIGNAL (documentActivated (KDevelop::IDocument*)),
            this,
            SLOT (documentActivated (KDevelop::IDocument*)));

   connect (m_tree,
            SIGNAL (itemDoubleClicked (QTreeWidgetItem*, int)),
            this,
            SLOT (treeDoubleClick (QTreeWidgetItem*, int)));

   m_jumpToSource = new KAction (this);
   m_jumpToSource->setText (i18n ("Jump to Source"));
   connect (m_jumpToSource, SIGNAL (triggered (bool)), this, SLOT (jumpToSource()));

   m_jumpToDeclaration = new KAction (this);
   m_jumpToDeclaration->setText (i18n ("Jump to Declaration"));
   m_jumpToDeclaration->setIcon (KIcon ("go-jump-declaration"));
   connect (m_jumpToDeclaration, SIGNAL (triggered (bool)), this, SLOT (jumpToDeclaration()));

   m_jumpToDefinition = new KAction (this);
   m_jumpToDefinition->setText (i18n ("Jump to Definition"));
   m_jumpToDefinition->setIcon (KIcon ("go-jump-definition"));
   connect (m_jumpToDefinition, SIGNAL (triggered (bool)), this, SLOT (jumpToDefinition()));

   m_jumpToInterface = new KAction (this);
   m_jumpToInterface->setText (i18n ("Jump to Interface"));
   connect (m_jumpToInterface, SIGNAL (triggered (bool)), this, SLOT (jumpToInterface()));

   m_jumpToImplementation = new KAction (this);
   m_jumpToImplementation->setText (i18n ("Jump to Implementation"));
   connect (m_jumpToImplementation, SIGNAL (triggered (bool)), this, SLOT (jumpToImplementation()));
}

NavigatorWidget::~NavigatorWidget()
{
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::contextMenuEvent (QContextMenuEvent* e)
{
   QMenu *menu = new QMenu (this);

   menu->addAction (m_jumpToSource);
   menu->addAction (m_jumpToDeclaration);
   menu->addAction (m_jumpToDefinition);
   menu->addAction (m_jumpToInterface);
   menu->addAction (m_jumpToImplementation);

   if (!menu->actions().isEmpty())
      menu->exec (QCursor::pos());
}

void NavigatorWidget::treeDoubleClick (QTreeWidgetItem* item, int column)
{
   trace ("tree double click");
   jumpToSource ();
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::jumpToSource ()
{
   trace_open ("jump to source");

   NavigatorItem * item = dynamic_cast <NavigatorItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
      KDevelop::DUContext * context = item->m_cachedContext.data();

      if (context != NULL)
      {
         KDevelop::IndexedString u = context->url();
         KDevelop::CursorInRevision c = context->range().start;
         if (c.isValid ())
         {
            KTextEditor::Cursor t = context->transformFromLocalRevision (c).textCursor();
            if (t.isValid ())
            {
               lock.unlock();
               KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), t);
            }
         }
      }

   }
   trace_close ();
   // from kdevplatform-1.2.0/plugins/quickopen/quickopenplugin.cpp, QuickOpenPlugin::jumpToNearestFunction
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::jumpToDeclaration ()
{
   trace_open ("jump to declaration");

   NavigatorItem * item = dynamic_cast <NavigatorItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
      KDevelop::Declaration* decl = item->m_cachedDecl.data();

      if (decl != NULL)
      {
         KDevelop::IndexedString u = decl->url();
         KDevelop::SimpleCursor c = decl->rangeInCurrentRevision().start;
         lock.unlock();
         KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), c.textCursor());
      }
   }

   trace_close ();
   // from kdevplatform-1.2.0/plugins/quickopen/quickopenplugin.cpp, QuickOpenPlugin::quickOpenDeclaration
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::jumpToDefinition ()
{
   trace_open ("jump to definition");

   NavigatorItem * item = dynamic_cast <NavigatorItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
      KDevelop::Declaration* decl = item->m_cachedDecl.data();

      if (decl != NULL)
      {
         KDevelop::IndexedString u = decl->url();
         KDevelop::SimpleCursor c = decl->rangeInCurrentRevision().start;
         KDevelop::FunctionDefinition * def = KDevelop::FunctionDefinition::definition (decl);
         {
            if (def != NULL)
            {
               def->activateSpecialization();
               u = def->url();
               c = def->rangeInCurrentRevision().start;
            }
         }
         lock.unlock();
         KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), c.textCursor());
      }
   }

   trace_close ();
   // from kdevplatform-1.2.0/plugins/quickopen/quickopenplugin.cpp, QuickOpenPlugin::quickOpenDefinition
}

/* ----------------------------------------------------------------------- */

KDevelop::Declaration * NavigatorWidget::searchDecl (KDevelop::DUContext * context, const QString id)
{
   KDevelop::Declaration * result = NULL;
   // trace_open ("searching in " + context->scopeIdentifier (true).toString ());

   foreach (KDevelop::Declaration* decl, context->localDeclarations())
   {
      trace (decl->qualifiedIdentifier().toString());
      if (decl->qualifiedIdentifier().toString () == id)
      {
         // trace_alert ("FOUND");
         result=decl;
      }
   }

   foreach (KDevelop::DUContext * child, context->childContexts())
   {
      if (result == NULL)
      {
         #if 0
         if (child->scopeIdentifier (true).toString () == id)
         {
            trace ("FOUND");
            result=child->owner();
         }
         #endif

         KDevelop::DUContext::ContextType type = child->type ();
         if (type == KDevelop::DUContext::Global ||
             type == KDevelop::DUContext::Namespace ||
             type == KDevelop::DUContext::Class)
         {
            result = searchDecl (child, id);
         }
      }
   }
   trace_close ();
   return result;
}

KDevelop::Declaration * NavigatorWidget::searchDeclaration (const QString id)
{
   KDevelop::Declaration * decl = NULL;
   QList <TopDUContext*> list = DUChain::self()->allChains();
   foreach (TopDUContext * context, list)
   {
      if (decl == NULL)
         decl = searchDecl (context, id);
   }
   return decl;
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::jumpToInterface ()
{
   trace ("jump to interface");
   NavigatorItem * item = dynamic_cast <NavigatorItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      QString id = item->data (0, Qt::UserRole).toString ();
      id = "Implementation::" + id;
      if (id != "")
      {
         KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
         KDevelop::Declaration * decl = searchDeclaration (id);
         if (decl != NULL)
         {
            trace ("jump to " + id);
            KDevelop::IndexedString u = decl->url();
            KDevelop::SimpleCursor c = decl->rangeInCurrentRevision().start;
            lock.unlock();
            KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), c.textCursor());
         }
      }
   }
   trace_close ();
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::jumpToImplementation ()
{
   trace_open ("jump to implementation");
   NavigatorItem * item = dynamic_cast <NavigatorItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      QString id = item->data (0, Qt::UserRole).toString ();
      id = "Implementation::" + id;
      if (id != "")
      {
         KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
         KDevelop::Declaration * decl = searchDeclaration (id);
         if (decl != NULL)
         {
            trace ("jump to " + id);
            KDevelop::IndexedString u = decl->url();
            KDevelop::SimpleCursor c = decl->rangeInCurrentRevision().start;

            KDevelop::FunctionDefinition * def = KDevelop::FunctionDefinition::definition (decl);
            {
               if (def != NULL)
               {
                  def->activateSpecialization();
                  u = def->url();
                  c = def->rangeInCurrentRevision().start;
               }
            }

            lock.unlock();
            KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), c.textCursor());
         }
      }
   }
   trace_close ();
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::show (QTreeWidgetItem * branch, KDevelop::DUContext * context)
{
   KDevelop::Declaration * decl = context->owner ();
   if (decl != NULL)
   {
      NavigatorItem * item  = new NavigatorItem ();

      if (branch == NULL)
         m_tree->addTopLevelItem (item);
      else
         branch->addChild (item);

      item->m_cachedContext = context;
      item->m_cachedDecl = decl;

      item->setIcon (0, KDevelop::DUChainUtils::iconForDeclaration (decl));
      item->setText (0, decl->qualifiedIdentifier().toString());
      item->setToolTip (0, decl->comment());
      // item->setStatusTip (0, decl->toString());

       item->setData (0, Qt::UserRole, decl->qualifiedIdentifier().toString());

      KDevelop::DUContext::ContextType type = context->type ();
      if (type == KDevelop::DUContext::Global ||
          type == KDevelop::DUContext::Namespace ||
          type == KDevelop::DUContext::Class)
      {
         foreach (KDevelop::DUContext * child, context->childContexts())
            show (item, child);
      }
   }
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::display ()
{
   m_tree->clear ();

   KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());

   KDevelop::IDocument* doc = KDevelop::ICore::self()->documentController()->activeDocument();
   if (doc && doc->textDocument() && doc->textDocument()->activeView())
   {
      KDevelop::TopDUContext* topContext = KDevelop::DUChainUtils::standardContextForUrl (doc->url());

      if (topContext != NULL)
      {
         foreach (KDevelop::DUContext * child, topContext->childContexts())
            show (NULL, child);
      }
   }

   lock.unlock ();
}

/* ----------------------------------------------------------------------- */

void NavigatorWidget::documentActivated (KDevelop::IDocument * doc)
{
   display ();
}

void NavigatorWidget::documentClosed (KDevelop::IDocument * doc)
{
   display ();
}

/* ----------------------------------------------------------------------- */

} // end of namespace

#include "navigator.moc"
