
/* review.cc */

#include "review.h"
#include "overview.h"

#include <QLabel>
#include <QHBoxLayout>
#include <QVBoxLayout>
#include <QHeaderView>
#include <QMenu>

#include <KLocalizedString>
#include <KIconLoader>

#include <ktexteditor/document.h>

#include <interfaces/icore.h>
#include <interfaces/idocumentcontroller.h>

#include <language/duchain/duchainutils.h>
#include <language/duchain/duchainlock.h>
#include <language/duchain/duchain.h>
#include <language/duchain/declaration.h>
#include <language/duchain/functiondefinition.h>

#include <iostream>
#include <typeinfo>
#include "trace.h"

// using namespace KDevelop;

/* ----------------------------------------------------------------------- */

namespace OverviewModule
{

/* ----------------------------------------------------------------------- */

ReviewWidget::ReviewWidget (QWidget* parent, OverviewPlugin* plugin) :
      QWidget (parent),
      m_plugin (plugin),
      m_tree (new QTreeWidget (this)),
      m_jumpToSource (NULL)
{
   setObjectName ("Review");
   setWindowTitle (i18n ("Review"));
   setWindowIcon (SmallIcon ("code-class"));
   setWhatsThis (i18n ("Review - DUChain tree viewer from Overview plugin"));

   showHeader ();

   QVBoxLayout* vbox = new QVBoxLayout (this);
   vbox->setMargin (0);
   vbox->addWidget (m_tree);
   setLayout (vbox);

   connect (KDevelop::ICore::self()->documentController(),
            SIGNAL (documentClosed (KDevelop::IDocument*)),
            this,
            SLOT (documentClosed (KDevelop::IDocument*)));

   connect (KDevelop::ICore::self()->documentController(),
            SIGNAL (documentActivated (KDevelop::IDocument*)),
            this,
            SLOT (documentActivated (KDevelop::IDocument*)));

   connect (m_tree,
            SIGNAL (itemDoubleClicked (QTreeWidgetItem*, int)),
            this,
            SLOT (treeDoubleClick (QTreeWidgetItem*, int)));

   connect (m_tree,
            SIGNAL (itemClicked (QTreeWidgetItem*, int)),
            this,
            SLOT (treeClick (QTreeWidgetItem*, int)));

   m_display = new KAction (this);
   m_display->setText (i18n ("Refresh"));
   connect (m_display, SIGNAL (triggered (bool)), this, SLOT (display()));

   m_jumpToSource = new KAction (this);
   m_jumpToSource->setText (i18n ("Jump to Source"));
   connect (m_jumpToSource, SIGNAL (triggered (bool)), this, SLOT (jumpToSource()));
}

ReviewWidget::~ReviewWidget()
{
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::contextMenuEvent (QContextMenuEvent* e)
{
   QMenu *menu = new QMenu (this);

   menu->addAction (m_jumpToSource);
   menu->addAction (m_display);

   menu->exec (QCursor::pos());
}

void ReviewWidget::treeClick (QTreeWidgetItem* item, int column)
{
   selectRange ();
}

void ReviewWidget::treeDoubleClick (QTreeWidgetItem* item, int column)
{
   jumpToSource ();
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::selectRange ()
{
   trace ("select range");

   ReviewItem * item = dynamic_cast <ReviewItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
      KDevelop::DUContext * context = item->m_cachedContext.data();
      if (context != NULL)
      {
         KDevelop::IndexedString u = context->url();
         KDevelop::RangeInRevision r = context->range();
         if (r.isValid ())
         {
             KTextEditor::Range t = context->transformFromLocalRevision (r).textRange();
             if (t.isValid ())
             {
                lock.unlock();
                KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), t);
             }
         }
      }
   }
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::jumpToSource ()
{
   trace ("jump to source");

   ReviewItem * item = dynamic_cast <ReviewItem *> (m_tree->currentItem ());
   if (item != NULL)
   {
      KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());
      KDevelop::DUContext * context = item->m_cachedContext.data();
      if (context != NULL)
      {
         KDevelop::IndexedString u = context->url();
         KDevelop::CursorInRevision c = context->range().start;
         if (c.isValid ())
         {
             KTextEditor::Cursor t = context->transformFromLocalRevision (c).textCursor();
             if (t.isValid ())
             {
                lock.unlock();
                KDevelop::ICore::self()->documentController()->openDocument (KUrl (u.str()), t);
             }
         }
      }
   }
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::showHeader ()
{
   m_tree->header()->setResizeMode (0, QHeaderView::ResizeToContents);
   // m_tree->header()->setStretchLastSection (false);
   m_tree->setColumnCount (6);

   QStringList labels;
   labels << "context->localScopeIdentifier()"
          << "context->type()"
          << "decl->identifier()"
          << "decl->qualifiedIdentifier()"
          << "decl->toString()"
          << "decl->comment()";
   m_tree->setHeaderLabels (labels);
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::show (QTreeWidgetItem * branch, KDevelop::DUContext * context)
{
   ReviewItem * item = new ReviewItem ();
   if (branch == NULL)
      m_tree->addTopLevelItem (item);
   else
      branch->addChild (item);

   item->setText (0, context->localScopeIdentifier().toString());
   item->m_cachedContext = context;

   {
      KDevelop::DUContext::ContextType type = context->type ();
      QString s = "";
      if (type == KDevelop::DUContext::Global)
         s = "Global";
      else if (type == KDevelop::DUContext::Namespace)
         s = "Namespace";
      else if (type == KDevelop::DUContext::Class)
         s = "Class";
      else if (type == KDevelop::DUContext::Function)
         s = "Function";
      else if (type == KDevelop::DUContext::Template)
         s = "Template";
      else if (type == KDevelop::DUContext::Enum)
         s = "Enum";
      else if (type == KDevelop::DUContext::Helper)
         s = "Helper";
      else if (type == KDevelop::DUContext::Other)
         s = "Other";
      item->setText (1, s);
    }

   KDevelop::Declaration * decl = context->owner ();
   if (decl != NULL)
   {
      item->setIcon (0, KDevelop::DUChainUtils::iconForDeclaration (decl));
      item->setText (2, decl->identifier().toString());
      item->setText (3, decl->qualifiedIdentifier().toString());
      item->setText (4, decl->toString());
      item->setText (5, decl->comment());
      item->setToolTip (0, decl->comment());
      item->m_cachedDecl = decl;
   }

   KDevelop::DUContext::ContextType type = context->type ();
   if (type == KDevelop::DUContext::Global ||
       type == KDevelop::DUContext::Namespace ||
       type == KDevelop::DUContext::Class)
   {
      foreach (KDevelop::DUContext * child, context->childContexts())
         show (item, child);
   }
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::display ()
{
   m_tree->clear ();

   KDevelop::DUChainReadLocker lock (KDevelop::DUChain::lock());

   KDevelop::IDocument* doc = KDevelop::ICore::self()->documentController()->activeDocument();
   if (doc && doc->textDocument() && doc->textDocument()->activeView())
   {
      KDevelop::TopDUContext* topContext = KDevelop::DUChainUtils::standardContextForUrl (doc->url());

      if (topContext != NULL)
      {
         foreach (KDevelop::DUContext * child, topContext->childContexts())
            show (NULL, child);
      }
   }

   lock.unlock ();
}

/* ----------------------------------------------------------------------- */

void ReviewWidget::documentActivated (KDevelop::IDocument * doc)
{
   display ();
}

void ReviewWidget::documentClosed (KDevelop::IDocument * doc)
{
   display ();
}

/* ----------------------------------------------------------------------- */

} // end of namespace

#include "review.moc"
